/*
 * COPYRIGHT:        See COPYRIGHT.TXT
 * PROJECT:          Ext2 File System Driver for WinNT/2K/XP
 * FILE:             devctl.c
 * PROGRAMMER:       Matt Wu <mattwu@163.com>
 * HOMEPAGE:         http://ext2.yeah.net
 * UPDATE HISTORY: 
 */

/* INCLUDES *****************************************************************/

#include "ext2fs.h"

/* GLOBALS ***************************************************************/

extern PEXT2_GLOBAL Ext2Global;

/* DEFINITIONS *************************************************************/

NTSTATUS
Ext2DeviceControlCompletion (
    IN PDEVICE_OBJECT   DeviceObject,
    IN PIRP             Irp,
    IN PVOID            Context);


#ifdef ALLOC_PRAGMA
//#pragma alloc_text(PAGE, Ext2DeviceControlCompletion)
#pragma alloc_text(PAGE, Ext2DeviceControl)
#pragma alloc_text(PAGE, Ext2DeviceControlNormal)
#if EXT2_UNLOAD
#pragma alloc_text(PAGE, Ext2PrepareToUnload)
#endif
#endif


NTSTATUS
Ext2DeviceControlCompletion (
    IN PDEVICE_OBJECT   DeviceObject,
    IN PIRP             Irp,
    IN PVOID            Context
    )
{
    if (Irp->PendingReturned) {
        IoMarkIrpPending(Irp);
    }

    return STATUS_SUCCESS;
}


NTSTATUS
Ext2DeviceControlNormal (IN PEXT2_IRP_CONTEXT IrpContext)
{
    PDEVICE_OBJECT  DeviceObject;
    BOOLEAN         CompleteRequest = TRUE;
    NTSTATUS        Status = STATUS_UNSUCCESSFUL;

    PEXT2_VCB       Vcb;

    PIRP            Irp;
    PIO_STACK_LOCATION IrpSp;
    PIO_STACK_LOCATION NextIrpSp;

    PDEVICE_OBJECT  TargetDeviceObject;
    
    __try {

        ASSERT(IrpContext != NULL);
        
        ASSERT((IrpContext->Identifier.Type == EXT2ICX) &&
            (IrpContext->Identifier.Size == sizeof(EXT2_IRP_CONTEXT)));
        
        CompleteRequest = TRUE;

        DeviceObject = IrpContext->DeviceObject;
    
        if (DeviceObject == Ext2Global->DeviceObject)  {
            Status = STATUS_INVALID_DEVICE_REQUEST;
            __leave;
        }
        
        Irp = IrpContext->Irp;
        IrpSp = IoGetCurrentIrpStackLocation(Irp);

        Vcb = (PEXT2_VCB) IrpSp->FileObject->FsContext;

        if (!((Vcb) && (Vcb->Identifier.Type == EXT2VCB) &&
              (Vcb->Identifier.Size == sizeof(EXT2_VCB)))) {
            Status = STATUS_INVALID_PARAMETER;
            __leave;
        }
        
        TargetDeviceObject = Vcb->TargetDeviceObject;
        
        //
        // Pass on the IOCTL to the driver below
        //
        
        CompleteRequest = FALSE;

        NextIrpSp = IoGetNextIrpStackLocation( Irp );
        *NextIrpSp = *IrpSp;

        IoSetCompletionRoutine(
            Irp,
            Ext2DeviceControlCompletion,
            NULL,
            FALSE,
            TRUE,
            TRUE );
        
        Status = IoCallDriver(TargetDeviceObject, Irp);

    } __finally  {

        if (!IrpContext->ExceptionInProgress) {
            if (IrpContext) {
                if (!CompleteRequest) {
                    IrpContext->Irp = NULL;
                }

                Ext2CompleteIrpContext(IrpContext, Status);
            }
        }
    }
    
    return Status;
}


#if EXT2_UNLOAD

NTSTATUS
Ext2PrepareToUnload (IN PEXT2_IRP_CONTEXT IrpContext)
{
    PDEVICE_OBJECT  DeviceObject;
    NTSTATUS        Status = STATUS_UNSUCCESSFUL;
    BOOLEAN         GlobalDataResourceAcquired = FALSE;
    
    __try {

        ASSERT(IrpContext != NULL);
        
        ASSERT((IrpContext->Identifier.Type == EXT2ICX) &&
            (IrpContext->Identifier.Size == sizeof(EXT2_IRP_CONTEXT)));
        
        DeviceObject = IrpContext->DeviceObject;
        
        if (DeviceObject != Ext2Global->DeviceObject) {
            Status = STATUS_INVALID_DEVICE_REQUEST;
            __leave;
        }
        
        ExAcquireResourceExclusiveLite(
            &Ext2Global->Resource,
            TRUE );
        
        GlobalDataResourceAcquired = TRUE;
        
        if (FlagOn(Ext2Global->Flags, EXT2_UNLOAD_PENDING)) {
            Ext2Print((DBG_ERROR, "Ext2PrepareUnload:  Already ready to unload.\n"));
            
            Status = STATUS_ACCESS_DENIED;
            
            __leave;
        }
        
        {
            PEXT2_VCB               Vcb;
            PLIST_ENTRY             ListEntry;

            ListEntry = Ext2Global->VcbList.Flink;

            while (ListEntry != &(Ext2Global->VcbList)) {

                Vcb = CONTAINING_RECORD(ListEntry, EXT2_VCB, Next);
                ListEntry = ListEntry->Flink;

                if (Vcb && (!Vcb->ReferenceCount) &&
                    IsFlagOn(Vcb->Flags, VCB_DISMOUNT_PENDING)) {
                    Ext2RemoveVcb(Vcb);
                    Ext2ClearVpbFlag(Vcb->Vpb, VPB_MOUNTED);

                    Ext2FreeVcb(Vcb);
                }
            }
        }

        if (!IsListEmpty(&(Ext2Global->VcbList))) {

            Ext2Print((DBG_ERROR, "Ext2PrepareUnload:  Mounted volumes exists.\n"));
            
            Status = STATUS_ACCESS_DENIED;
            
            __leave;
        }
        
        IoUnregisterFileSystem(Ext2Global->DeviceObject);
        
        Ext2Global->DriverObject->DriverUnload = DriverUnload;
        
        SetFlag(Ext2Global->Flags ,EXT2_UNLOAD_PENDING);
        
        Ext2Print((DBG_INFO, "Ext2PrepareToUnload: Driver is ready to unload.\n"));
        
        Status = STATUS_SUCCESS;

    } __finally {

        if (GlobalDataResourceAcquired) {
            ExReleaseResourceLite(&Ext2Global->Resource);
        }
        
        if (!IrpContext->ExceptionInProgress) {
            Ext2CompleteIrpContext(IrpContext, Status);
        }
    }
    
    return Status;
}

#endif

NTSTATUS
Ext2FilesystemProperty(
    IN PEXT2_IRP_CONTEXT IrpContext,
    IN  PDEVICE_OBJECT  DeviceObject,
    IN  PEXT2_VOLUME_PROPERTY Property,
    IN  ULONG           Length
    )
{
    NTSTATUS        Status = STATUS_UNSUCCESSFUL;
    BOOLEAN         GlobalDataResourceAcquired = FALSE;
    BOOLEAN         VcbResourceAcquired = FALSE;
    struct nls_table * PageTable = NULL;
    PEXT2_VCB       Vcb = NULL;
    
    __try {

        ASSERT(IrpContext != NULL);
        
        ASSERT((IrpContext->Identifier.Type == EXT2ICX) &&
            (IrpContext->Identifier.Size == sizeof(EXT2_IRP_CONTEXT)));
        
        DeviceObject = IrpContext->DeviceObject;
        
        if (DeviceObject == Ext2Global->DeviceObject) {

            if (Property->Command != APP_CMD_SET_PROPERTY) {
                Status = STATUS_INVALID_PARAMETER;
                __leave;
            }

            if (!IsFlagOn(Property->Flags, EXT2_FLAG_VP_SET_GLOBAL)) {
                Status = STATUS_INVALID_PARAMETER;
                __leave;
            }

            ExAcquireResourceExclusiveLite(
                &Ext2Global->Resource,
                TRUE );
            GlobalDataResourceAcquired = TRUE;

            if (Property->bReadonly) {
                ClearFlag(Ext2Global->Flags, EXT2_SUPPORT_WRITING);
                ClearFlag(Ext2Global->Flags, EXT3_FORCE_WRITING);
            } else {
                SetFlag(Ext2Global->Flags, EXT2_SUPPORT_WRITING);
                if (Property->bExt3Writable) {
                    SetFlag(Ext2Global->Flags, EXT3_FORCE_WRITING);
                } else {
                    ClearFlag(Ext2Global->Flags, EXT3_FORCE_WRITING);
                }
            }

            PageTable = load_nls(Property->Codepage);
            if (PageTable) {
                memset(Ext2Global->Codepage.UniName, 0, sizeof(WCHAR) * CODEPAGE_MAXLEN);
                memcpy(Ext2Global->Codepage.AnsiName, Property->Codepage, CODEPAGE_MAXLEN);
                Ext2Global->Codepage.PageTable = PageTable;
            } else {
                DbgBreak();
            }

        } else {

            Vcb = (PEXT2_VCB) DeviceObject->DeviceExtension;

            if (!((Vcb) && (Vcb->Identifier.Type == EXT2VCB) &&
                  (Vcb->Identifier.Size == sizeof(EXT2_VCB)))) {
                Status = STATUS_INVALID_PARAMETER;
                __leave;
            }

            ExAcquireResourceExclusiveLite(
                &Vcb->MainResource,
                TRUE );
            VcbResourceAcquired = TRUE;

            if (Property->Command == APP_CMD_SET_PROPERTY) {

                if (Property->bReadonly) {
                    SetFlag(Vcb->Flags, VCB_READ_ONLY);
                } else {
                    ClearFlag(Vcb->Flags, VCB_READ_ONLY);
                }

                PageTable = load_nls(Property->Codepage);
                if (PageTable) {
                    Vcb->PageTable = PageTable;
                    Ext2InitializeLabel(Vcb, Vcb->SuperBlock);
                } else {
                    DbgBreak();
                }

            } else if (Property->Command == APP_CMD_QUERY_PROPERTY) {

                Property->bExt2 = TRUE;
                Property->bExt3 = Vcb->IsExt3fs;
                Property->bReadonly = IsFlagOn(Vcb->Flags, VCB_READ_ONLY);
                if (Property->bReadonly) {
                    Property->bExt3Writable = FALSE;
                } else {
                    Property->bExt3Writable = TRUE;
                }

                RtlZeroMemory(Property->Codepage, CODE_PAGE_LEN);
                if (Vcb->PageTable) {
                    strcpy(Property->Codepage, Vcb->PageTable->charset);
                } else {
                    strcpy(Property->Codepage, "default");
                }

            } else {
                Status = STATUS_INVALID_PARAMETER;
                __leave;
            }
        }

        IrpContext->Irp->IoStatus.Information = Length;
        Status = STATUS_SUCCESS;

    } __finally {

        if (VcbResourceAcquired) {
            ExReleaseResourceLite(&Vcb->MainResource);
        }
        
        if (GlobalDataResourceAcquired) {
            ExReleaseResourceLite(&Ext2Global->Resource);
        }

        if (!IrpContext->ExceptionInProgress) {
            Ext2CompleteIrpContext(IrpContext, Status);
        }
    }
    
    return Status;
}

NTSTATUS
Ext2DeviceControl (IN PEXT2_IRP_CONTEXT IrpContext)
{
    PIRP                Irp;
    PIO_STACK_LOCATION  irpSp;
    ULONG               code;
    ULONG               length;
    NTSTATUS            Status;
    
    ASSERT(IrpContext);
    
    ASSERT((IrpContext->Identifier.Type == EXT2ICX) &&
        (IrpContext->Identifier.Size == sizeof(EXT2_IRP_CONTEXT)));
    
    Irp = IrpContext->Irp;
    
    irpSp = IoGetCurrentIrpStackLocation(Irp);
    
    code = irpSp->Parameters.DeviceIoControl.IoControlCode;
    length = irpSp->Parameters.DeviceIoControl.OutputBufferLength;
    
    switch (code) {

    case IOCTL_APP_VOLUME_PROPERTY:
        Status = Ext2FilesystemProperty(
                        IrpContext, IrpContext->DeviceObject,
                        Irp->AssociatedIrp.SystemBuffer,
                        length
                        );
         break;

#if EXT2_UNLOAD
    case IOCTL_PREPARE_TO_UNLOAD:
        Status = Ext2PrepareToUnload(IrpContext);
        break;
#endif
    default:
        Status = Ext2DeviceControlNormal(IrpContext);
    }
    
    return Status;
}
