/*
 * COPYRIGHT:        See COPYRIGHT.TXT
 * PROJECT:          Ext2 File System Driver for WinNT/2K/XP
 * FILE:             fileinfo.c
 * PROGRAMMER:       Matt Wu <mattwu@163.com>
 * HOMEPAGE:         http://ext2.yeah.net
 * UPDATE HISTORY: 
 */

/* INCLUDES *****************************************************************/

#include "ext2fs.h"

/* GLOBALS ***************************************************************/

extern PEXT2_GLOBAL Ext2Global;

/* DEFINITIONS *************************************************************/

#ifdef ALLOC_PRAGMA
#pragma alloc_text(PAGE, Ext2QueryInformation)
#pragma alloc_text(PAGE, Ext2SetInformation)
#pragma alloc_text(PAGE, Ext2ExpandFile)
#pragma alloc_text(PAGE, Ext2TruncateFile)
#pragma alloc_text(PAGE, Ext2SetDispositionInfo)
#pragma alloc_text(PAGE, Ext2SetRenameInfo)
#pragma alloc_text(PAGE, Ext2DeleteFile)
#endif


NTSTATUS
Ext2QueryInformation (IN PEXT2_IRP_CONTEXT IrpContext)
{
    PDEVICE_OBJECT          DeviceObject;
    NTSTATUS                Status = STATUS_UNSUCCESSFUL;
    PFILE_OBJECT            FileObject;
    PEXT2_VCB               Vcb;
    PEXT2_FCB               Fcb;
    PEXT2_CCB               Ccb;
    PIRP                    Irp;
    PIO_STACK_LOCATION      IoStackLocation;
    FILE_INFORMATION_CLASS  FileInformationClass;
    ULONG                   Length;
    PVOID                   Buffer;
    BOOLEAN                 FcbResourceAcquired = FALSE;
    
    __try {

        ASSERT(IrpContext != NULL);
        
        ASSERT((IrpContext->Identifier.Type == EXT2ICX) &&
            (IrpContext->Identifier.Size == sizeof(EXT2_IRP_CONTEXT)));
        
        DeviceObject = IrpContext->DeviceObject;
        
        //
        // This request is not allowed on the main device object
        //
        if (DeviceObject == Ext2Global->DeviceObject) {
            Status = STATUS_INVALID_DEVICE_REQUEST;
            __leave;
        }
        
        FileObject = IrpContext->FileObject;
        
        Fcb = (PEXT2_FCB) FileObject->FsContext;
        
        if (Fcb == NULL) {
            Status = STATUS_INVALID_PARAMETER;
            __leave;
        }
        
        //
        // This request is not allowed on volumes
        //
        if (Fcb->Identifier.Type == EXT2VCB) {
            Status = STATUS_INVALID_PARAMETER;
            __leave;
        }
        
        ASSERT((Fcb->Identifier.Type == EXT2FCB) &&
            (Fcb->Identifier.Size == sizeof(EXT2_FCB)));

        Vcb = Fcb->Vcb;

        {
            if (!ExAcquireResourceSharedLite(
                &Fcb->MainResource,
                IsFlagOn(IrpContext->Flags, IRP_CONTEXT_FLAG_WAIT)
                )) {

                Status = STATUS_PENDING;
                __leave;
            }
            
            FcbResourceAcquired = TRUE;
        }
        
        Ccb = (PEXT2_CCB) FileObject->FsContext2;
        
        ASSERT(Ccb != NULL);
        
        ASSERT((Ccb->Identifier.Type == EXT2CCB) &&
            (Ccb->Identifier.Size == sizeof(EXT2_CCB)));

        Irp = IrpContext->Irp;
        IoStackLocation = IoGetCurrentIrpStackLocation(Irp);
        FileInformationClass =
            IoStackLocation->Parameters.QueryFile.FileInformationClass;

        Length = IoStackLocation->Parameters.QueryFile.Length;
        Buffer = Irp->AssociatedIrp.SystemBuffer;
        RtlZeroMemory(Buffer, Length);
      
        switch (FileInformationClass) {

            case FileBasicInformation:
            {
                PFILE_BASIC_INFORMATION FileBasicInformation;

                if (Length < sizeof(FILE_BASIC_INFORMATION)) {
                    Status = STATUS_INFO_LENGTH_MISMATCH;
                    __leave;
                }

                FileBasicInformation = (PFILE_BASIC_INFORMATION) Buffer;

                if (IsRoot(Fcb)) {
                    FileBasicInformation->CreationTime = 
                    FileBasicInformation->LastAccessTime =
                    FileBasicInformation->LastWriteTime =
                    FileBasicInformation->ChangeTime = Ext2NtTime(0);
                } else {
                    FileBasicInformation->CreationTime = Fcb->Ext2Mcb->CreationTime;
                    FileBasicInformation->LastAccessTime = Fcb->Ext2Mcb->LastAccessTime;
                    FileBasicInformation->LastWriteTime = Fcb->Ext2Mcb->LastWriteTime;
                    FileBasicInformation->ChangeTime = Fcb->Ext2Mcb->ChangeTime;
                }

                FileBasicInformation->FileAttributes = Fcb->Ext2Mcb->FileAttr;
                if (FileBasicInformation->FileAttributes == 0) {
                    FileBasicInformation->FileAttributes = FILE_ATTRIBUTE_NORMAL;
                }

                Irp->IoStatus.Information = sizeof(FILE_BASIC_INFORMATION);
                Status = STATUS_SUCCESS;
            }
            break;

            case FileStandardInformation:
            {
                PFILE_STANDARD_INFORMATION FSI;
                
                if (Length < sizeof(FILE_STANDARD_INFORMATION)) {
                    Status = STATUS_INFO_LENGTH_MISMATCH;
                    __leave;
                }
                
                FSI = (PFILE_STANDARD_INFORMATION) Buffer;
                
                FSI->NumberOfLinks = Fcb->Inode->i_links_count;
                
                if (IsFlagOn(Fcb->Vcb->Flags, VCB_READ_ONLY))
                    FSI->DeletePending = FALSE;
                else
                    FSI->DeletePending = IsFlagOn(Fcb->Flags, FCB_DELETE_PENDING);                
                
                if (IsDirectory(Fcb)) {
                    FSI->Directory = TRUE;
                    FSI->AllocationSize.QuadPart = 0;
                    FSI->EndOfFile.QuadPart = 0;
                } else {
                    FSI->Directory = FALSE;
                    FSI->AllocationSize = Fcb->Header.AllocationSize;
                    FSI->EndOfFile = Fcb->Header.FileSize;
                }

                Irp->IoStatus.Information = sizeof(FILE_STANDARD_INFORMATION);
                Status = STATUS_SUCCESS;
            }
            break;

           
            case FileInternalInformation:
            {
                PFILE_INTERNAL_INFORMATION FileInternalInformation;
                
                if (Length < sizeof(FILE_INTERNAL_INFORMATION)) {
                    Status = STATUS_INFO_LENGTH_MISMATCH;
                    __leave;
                }
                
                FileInternalInformation = (PFILE_INTERNAL_INFORMATION) Buffer;
                
                /* we use the inode number as the internal index */
                FileInternalInformation->IndexNumber.QuadPart = (LONGLONG)Fcb->Ext2Mcb->Inode;
                
                Irp->IoStatus.Information = sizeof(FILE_INTERNAL_INFORMATION);
                Status = STATUS_SUCCESS;
            }
            break;


            case FileEaInformation:
            {
                PFILE_EA_INFORMATION FileEaInformation;
                
                if (Length < sizeof(FILE_EA_INFORMATION)) {
                    Status = STATUS_INFO_LENGTH_MISMATCH;
                    __leave;
                }
                
                FileEaInformation = (PFILE_EA_INFORMATION) Buffer;
                
                // Romfs doesn't have any extended attributes
                FileEaInformation->EaSize = 0;
                
                Irp->IoStatus.Information = sizeof(FILE_EA_INFORMATION);
                Status = STATUS_SUCCESS;
            }
            break;

            case FileNameInformation:
            {
                PFILE_NAME_INFORMATION FileNameInformation;
                
                if (Length < sizeof(FILE_NAME_INFORMATION) +
                    Fcb->Ext2Mcb->ShortName.Length - sizeof(WCHAR)) {
                    Status = STATUS_INFO_LENGTH_MISMATCH;
                    __leave;
                }
                
                FileNameInformation = (PFILE_NAME_INFORMATION) Buffer;
                
                FileNameInformation->FileNameLength = Fcb->Ext2Mcb->ShortName.Length;
                
                RtlCopyMemory(
                    FileNameInformation->FileName,
                    Fcb->Ext2Mcb->ShortName.Buffer,
                    Fcb->Ext2Mcb->ShortName.Length );
                
                Irp->IoStatus.Information = sizeof(FILE_NAME_INFORMATION) +
                    Fcb->Ext2Mcb->ShortName.Length - sizeof(WCHAR);
                Status = STATUS_SUCCESS;
            }
            break;
            
            case FilePositionInformation:
            {
                PFILE_POSITION_INFORMATION FilePositionInformation;
                
                if (Length < sizeof(FILE_POSITION_INFORMATION)) {
                    Status = STATUS_INFO_LENGTH_MISMATCH;
                    __leave;
                }
                
                FilePositionInformation = (PFILE_POSITION_INFORMATION) Buffer;
                
                FilePositionInformation->CurrentByteOffset =
                    FileObject->CurrentByteOffset;
                
                Irp->IoStatus.Information = sizeof(FILE_POSITION_INFORMATION);
                Status = STATUS_SUCCESS;
            }
            break;

            
            case FileAllInformation:
            {
                PFILE_ALL_INFORMATION       FileAllInformation;
                PFILE_BASIC_INFORMATION     FileBasicInformation;
                PFILE_STANDARD_INFORMATION  FSI;
                PFILE_INTERNAL_INFORMATION  FileInternalInformation;
                PFILE_EA_INFORMATION        FileEaInformation;
                PFILE_POSITION_INFORMATION  FilePositionInformation;
                PFILE_NAME_INFORMATION      FileNameInformation;
                
                if (Length < sizeof(FILE_ALL_INFORMATION)) {
                    Status = STATUS_INFO_LENGTH_MISMATCH;
                    __leave;
                }
                
                FileAllInformation = (PFILE_ALL_INFORMATION) Buffer;
                
                FileBasicInformation =
                    &FileAllInformation->BasicInformation;
                
                FSI =
                    &FileAllInformation->StandardInformation;
                
                FileInternalInformation =
                    &FileAllInformation->InternalInformation;
                
                FileEaInformation =
                    &FileAllInformation->EaInformation;
                
                FilePositionInformation =
                    &FileAllInformation->PositionInformation;
                
                FileNameInformation =
                    &FileAllInformation->NameInformation;
                
                if (IsRoot(Fcb)) {
                    FileBasicInformation->CreationTime =
                    FileBasicInformation->LastAccessTime =
                    FileBasicInformation->LastWriteTime =
                    FileBasicInformation->ChangeTime = Ext2NtTime(0);
                } else {
                    FileBasicInformation->CreationTime = Fcb->Ext2Mcb->CreationTime;
                    FileBasicInformation->LastAccessTime = Fcb->Ext2Mcb->LastAccessTime;
                    FileBasicInformation->LastWriteTime = Fcb->Ext2Mcb->LastWriteTime;
                    FileBasicInformation->ChangeTime = Fcb->Ext2Mcb->ChangeTime;
                }

                FileBasicInformation->FileAttributes = Fcb->Ext2Mcb->FileAttr;
                if (FileBasicInformation->FileAttributes == 0) {
                    FileBasicInformation->FileAttributes = FILE_ATTRIBUTE_NORMAL;
                }

                FSI->NumberOfLinks = Fcb->Inode->i_links_count;

                if (IsFlagOn(Fcb->Vcb->Flags, VCB_READ_ONLY))
                    FSI->DeletePending = FALSE;
                else
                    FSI->DeletePending = IsFlagOn(Fcb->Flags, FCB_DELETE_PENDING);

                if (IsDirectory(Fcb)) {
                    FSI->Directory = TRUE;
                    FSI->AllocationSize.QuadPart = 0;
                    FSI->EndOfFile.QuadPart = 0;
                } else {
                    FSI->Directory = FALSE;
                    FSI->AllocationSize = Fcb->Header.AllocationSize;
                    FSI->EndOfFile = Fcb->Header.FileSize;
                }

                // The "inode number"
                FileInternalInformation->IndexNumber.QuadPart = (LONGLONG)Fcb->Ext2Mcb->Inode;

                // Romfs doesn't have any extended attributes
                FileEaInformation->EaSize = 0;

                FilePositionInformation->CurrentByteOffset =
                    FileObject->CurrentByteOffset;
                
                if (Length < sizeof(FILE_ALL_INFORMATION) +
                    Fcb->Ext2Mcb->ShortName.Length - sizeof(WCHAR)) {
                    Irp->IoStatus.Information = sizeof(FILE_ALL_INFORMATION);
                    Status = STATUS_BUFFER_OVERFLOW;
                    __leave;
                }
                
                FileNameInformation->FileNameLength = Fcb->Ext2Mcb->ShortName.Length;
                
                RtlCopyMemory(
                    FileNameInformation->FileName,
                    Fcb->Ext2Mcb->ShortName.Buffer,
                    Fcb->Ext2Mcb->ShortName.Length
                    );
                
                Irp->IoStatus.Information = sizeof(FILE_ALL_INFORMATION) +
                    Fcb->Ext2Mcb->ShortName.Length - sizeof(WCHAR);
                Status = STATUS_SUCCESS;
            }
            break;
        
            /*
            case FileAlternateNameInformation:
            {
                // TODO: Handle FileAlternateNameInformation
        
                // Here we would like to use RtlGenerate8dot3Name but I don't
                // know how to use the argument PGENERATE_NAME_CONTEXT
            }
            */
          
            case FileNetworkOpenInformation:
            {
                PFILE_NETWORK_OPEN_INFORMATION PFNOI;
            
                if (Length < sizeof(FILE_NETWORK_OPEN_INFORMATION)) {
                    Status = STATUS_INFO_LENGTH_MISMATCH;
                    __leave;
                }

                PFNOI = (PFILE_NETWORK_OPEN_INFORMATION) Buffer;

                PFNOI->FileAttributes = Fcb->Ext2Mcb->FileAttr;
                if (PFNOI->FileAttributes == 0) {
                    PFNOI->FileAttributes = FILE_ATTRIBUTE_NORMAL;
                }

                if (IsRoot(Fcb)) {
                    PFNOI->CreationTime =
                    PFNOI->LastAccessTime =
                    PFNOI->LastWriteTime =
                    PFNOI->ChangeTime = Ext2NtTime(0);
                } else {
                    PFNOI->CreationTime   = Fcb->Ext2Mcb->CreationTime;
                    PFNOI->LastAccessTime = Fcb->Ext2Mcb->LastAccessTime;
                    PFNOI->LastWriteTime  = Fcb->Ext2Mcb->LastWriteTime;
                    PFNOI->ChangeTime     = Fcb->Ext2Mcb->ChangeTime;
                }

                if (IsDirectory(Fcb)) {
                    PFNOI->AllocationSize.QuadPart = 0;
                    PFNOI->EndOfFile.QuadPart = 0;
                } else {
                    PFNOI->AllocationSize = Fcb->Header.AllocationSize;
                    PFNOI->EndOfFile      = Fcb->Header.FileSize;
                }

                Irp->IoStatus.Information =
                    sizeof(FILE_NETWORK_OPEN_INFORMATION);
                Status = STATUS_SUCCESS;
            }
            break;

#if (_WIN32_WINNT >= 0x0500)

        case FileAttributeTagInformation:
            {
                PFILE_ATTRIBUTE_TAG_INFORMATION FATI;
                
                if (Length < sizeof(FILE_ATTRIBUTE_TAG_INFORMATION)) {
                    Status = STATUS_INFO_LENGTH_MISMATCH;
                    __leave;
                }
                
                FATI = (PFILE_ATTRIBUTE_TAG_INFORMATION) Buffer;
                
                FATI->FileAttributes = Fcb->Ext2Mcb->FileAttr;
                FATI->ReparseTag = IO_REPARSE_TAG_RESERVED_ZERO;
                
                Irp->IoStatus.Information = sizeof(FILE_ATTRIBUTE_TAG_INFORMATION);
                Status = STATUS_SUCCESS;
            }
            break;
#endif // (_WIN32_WINNT >= 0x0500)

        
        default:
            Ext2Print((DBG_ERROR, "Ext2QueryInformation: invalid class: %d\n",
                              FileInformationClass));
            Status = STATUS_INVALID_PARAMETER; /* STATUS_INVALID_INFO_CLASS; */
            break;
        }

    } __finally {

        if (FcbResourceAcquired) {
            ExReleaseResourceLite(&Fcb->MainResource);
        }

        if (!IrpContext->ExceptionInProgress) {
            if (Status == STATUS_PENDING) {
                Ext2QueueRequest(IrpContext);
            } else {
                Ext2CompleteIrpContext(IrpContext,  Status);
            }
        }
    }
    
    return Status;
}


NTSTATUS
Ext2SetInformation (IN PEXT2_IRP_CONTEXT IrpContext)
{
    PDEVICE_OBJECT          DeviceObject;
    NTSTATUS                Status = STATUS_UNSUCCESSFUL;
    PEXT2_VCB               Vcb;
    PFILE_OBJECT            FileObject;
    PEXT2_FCB               Fcb;
    PEXT2_CCB               Ccb;
    PIRP                    Irp;
    PIO_STACK_LOCATION      IoStackLocation;
    FILE_INFORMATION_CLASS  FileInformationClass;

    ULONG                   NotifyFilter = 0;

    ULONG                   Length;
    PVOID                   Buffer;

    BOOLEAN                 FcbMainResourceAcquired = FALSE;
    BOOLEAN                 FcbPagingIoResourceAcquired = FALSE;

    BOOLEAN                 CacheInitialized = FALSE;

    __try {

        ASSERT(IrpContext != NULL);
        
        ASSERT((IrpContext->Identifier.Type == EXT2ICX) &&
            (IrpContext->Identifier.Size == sizeof(EXT2_IRP_CONTEXT)));
        
        DeviceObject = IrpContext->DeviceObject;

        //
        // This request is not allowed on the main device object
        //
        if (DeviceObject == Ext2Global->DeviceObject) {
            Status = STATUS_INVALID_DEVICE_REQUEST;
            __leave;
        }
        
        Vcb = (PEXT2_VCB) DeviceObject->DeviceExtension;
        
        ASSERT(Vcb != NULL);
        
        ASSERT((Vcb->Identifier.Type == EXT2VCB) &&
            (Vcb->Identifier.Size == sizeof(EXT2_VCB)));

        ASSERT(IsMounted(Vcb));

        FileObject = IrpContext->FileObject;
        Fcb = (PEXT2_FCB) FileObject->FsContext;

        //
        // This request is not allowed on volumes
        //
        if (Fcb == NULL || Fcb->Identifier.Type == EXT2VCB) {
            Status = STATUS_INVALID_PARAMETER;
            __leave;
        }
        
        ASSERT((Fcb->Identifier.Type == EXT2FCB) &&
            (Fcb->Identifier.Size == sizeof(EXT2_FCB)));

        if (IsFlagOn(Fcb->Flags, FCB_FILE_DELETED)) {
            Status = STATUS_FILE_DELETED;
            __leave;
        }

        Ccb = (PEXT2_CCB) FileObject->FsContext2;
        
        ASSERT(Ccb != NULL);
        
        ASSERT((Ccb->Identifier.Type == EXT2CCB) &&
            (Ccb->Identifier.Size == sizeof(EXT2_CCB)));
        
        Irp = IrpContext->Irp;
        
        IoStackLocation = IoGetCurrentIrpStackLocation(Irp);
        
        FileInformationClass =
            IoStackLocation->Parameters.SetFile.FileInformationClass;
        
        Length = IoStackLocation->Parameters.SetFile.Length;
        
        Buffer = Irp->AssociatedIrp.SystemBuffer;

        if ( !IsDirectory(Fcb) && !FlagOn(Fcb->Flags, FCB_PAGE_FILE) &&
             ((FileInformationClass == FileEndOfFileInformation) ||
             (FileInformationClass == FileAllocationInformation))) {

            Status = FsRtlCheckOplock( &Fcb->Oplock,
                                       Irp,
                                       IrpContext,
                                       NULL,
                                       NULL );

            if (Status != STATUS_SUCCESS) {
                __leave;
            }

            //
            //  Set the flag indicating if Fast I/O is possible
            //

            Fcb->Header.IsFastIoPossible = Ext2IsFastIoPossible(Fcb);
        }

        if (!IsFlagOn(Fcb->Flags, FCB_PAGE_FILE)) {

            if (!ExAcquireResourceExclusiveLite(
                &Fcb->MainResource,
                IsFlagOn(IrpContext->Flags, IRP_CONTEXT_FLAG_WAIT) )) {
                Status = STATUS_PENDING;
                __leave;
            }
            
            FcbMainResourceAcquired = TRUE;
        }
        
        if (IsFlagOn(Vcb->Flags, VCB_READ_ONLY)) {

            if (FileInformationClass != FilePositionInformation) {
                Status = STATUS_MEDIA_WRITE_PROTECTED;
                __leave;
            }
        }

        if ( FileInformationClass == FileAllocationInformation ||
             FileInformationClass == FileEndOfFileInformation) {

            if (!ExAcquireResourceExclusiveLite(
                &Fcb->PagingIoResource,
                IsFlagOn(IrpContext->Flags, IRP_CONTEXT_FLAG_WAIT) )) {
                Status = STATUS_PENDING;
                DbgBreak();
                __leave;
            }
            
            FcbPagingIoResourceAcquired = TRUE;
        }
       
/*        
        if (FileInformationClass != FileDispositionInformation 
            && FlagOn(Fcb->Flags, FCB_DELETE_PENDING))
        {
            Status = STATUS_DELETE_PENDING;
            __leave;
        }
*/
        switch (FileInformationClass) {

        case FileBasicInformation:
            {
                PFILE_BASIC_INFORMATION FBI = (PFILE_BASIC_INFORMATION) Buffer;               
                PEXT2_INODE Ext2Inode = Fcb->Inode;

                if (FBI->CreationTime.QuadPart != 0 && FBI->CreationTime.QuadPart != -1) {
                    Fcb->Ext2Mcb->CreationTime = FBI->CreationTime;
                    Ext2Inode->i_ctime = Ext2LinuxTime(FBI->CreationTime);
                    NotifyFilter |= FILE_NOTIFY_CHANGE_CREATION;
                }

                if (FBI->LastAccessTime.QuadPart != 0 && FBI->LastAccessTime.QuadPart != -1) {
                    Fcb->Ext2Mcb->LastAccessTime = FBI->LastAccessTime;
                    Ext2Inode->i_atime = Ext2LinuxTime(FBI->LastAccessTime);
                    NotifyFilter |= FILE_NOTIFY_CHANGE_LAST_ACCESS;
                }

                if (FBI->LastWriteTime.QuadPart != 0 && FBI->LastWriteTime.QuadPart != -1) {
                    Fcb->Ext2Mcb->LastWriteTime = FBI->LastWriteTime;
                    Ext2Inode->i_mtime = Ext2LinuxTime(FBI->LastWriteTime);
                    NotifyFilter |= FILE_NOTIFY_CHANGE_LAST_WRITE;
                }

                if (FBI->ChangeTime.QuadPart !=0 && FBI->ChangeTime.QuadPart != -1) {
                    Fcb->Ext2Mcb->ChangeTime = FBI->ChangeTime;
                }

                if (FBI->FileAttributes != 0) {

                    NotifyFilter |= FILE_NOTIFY_CHANGE_ATTRIBUTES;

                    if (IsFlagOn(FBI->FileAttributes, FILE_ATTRIBUTE_READONLY)) {
                        Ext2SetReadOnly(Fcb->Inode->i_mode);
                        SetFlag(Fcb->Ext2Mcb->FileAttr, FILE_ATTRIBUTE_READONLY);
                    } else {
                        Ext2SetWritable(Fcb->Inode->i_mode);
                        ClearFlag(Fcb->Ext2Mcb->FileAttr, FILE_ATTRIBUTE_READONLY);
                    }

                    if (FBI->FileAttributes & FILE_ATTRIBUTE_TEMPORARY) {
                        SetFlag(FileObject->Flags, FO_TEMPORARY_FILE);
                    } else {
                        ClearFlag(FileObject->Flags, FO_TEMPORARY_FILE);
                    }
                }

                if (NotifyFilter != 0) {
                    if ( Ext2SaveInode( IrpContext, Vcb, 
                            Fcb->Ext2Mcb->Inode, Ext2Inode)) {
                        Status = STATUS_SUCCESS;
                    }
                }

                ClearFlag(NotifyFilter, FILE_NOTIFY_CHANGE_LAST_ACCESS);
                Status = STATUS_SUCCESS;
            }

            break;

        case FileAllocationInformation:
            {
                PFILE_ALLOCATION_INFORMATION FAI = (PFILE_ALLOCATION_INFORMATION)Buffer;

                if (IsDirectory(Fcb)) {
                    Status = STATUS_INVALID_DEVICE_REQUEST;
                    DbgBreak();
                    __leave;
                }

                if ((FileObject->SectionObjectPointer->DataSectionObject != NULL) &&
                    (FileObject->SectionObjectPointer->SharedCacheMap == NULL) &&
                    !FlagOn(Irp->Flags, IRP_PAGING_IO)) {

                    ASSERT( !FlagOn( FileObject->Flags, FO_CLEANUP_COMPLETE ) );

                    CcInitializeCacheMap( 
                            FileObject,
                            (PCC_FILE_SIZES)&(Fcb->Header.AllocationSize),
                            FALSE,
                            &(Ext2Global->CacheManagerNoOpCallbacks),
                            Fcb );

                    CacheInitialized = TRUE;
                }

                if (FAI->AllocationSize.QuadPart == Fcb->Header.AllocationSize.QuadPart) {

                    Status = STATUS_SUCCESS;

                } else if (FAI->AllocationSize.QuadPart > Fcb->Header.AllocationSize.QuadPart) {

                    Status = Ext2ExpandFile(
                                IrpContext,
                                Vcb, Fcb,
                                &(FAI->AllocationSize));

                    NotifyFilter = FILE_NOTIFY_CHANGE_SIZE;

                } else {

                    if (MmCanFileBeTruncated(&(Fcb->SectionObject), &(FAI->AllocationSize)))  {

                        LARGE_INTEGER EndOfFile;

                        EndOfFile.QuadPart = FAI->AllocationSize.QuadPart + (LONGLONG)(Vcb->BlockSize - 1);

                        Status = Ext2TruncateFile(IrpContext, Vcb, Fcb, &(EndOfFile));

                        NotifyFilter = FILE_NOTIFY_CHANGE_SIZE;

                        if (Fcb->Header.AllocationSize.QuadPart < Fcb->Header.FileSize.QuadPart) {
                            Fcb->Header.FileSize = Fcb->Ext2Mcb->FileSize = Fcb->Header.AllocationSize;
                            Fcb->Inode->i_size = Fcb->Header.FileSize.LowPart;
                            if (S_ISREG(Fcb->Inode->i_mode)) {
                                Fcb->Inode->i_size_high = (__u32)(Fcb->Header.FileSize.HighPart);
                            }
                            if (Fcb->Header.ValidDataLength.QuadPart > Fcb->Header.FileSize.QuadPart)
                                Fcb->Header.ValidDataLength.QuadPart = Fcb->Header.FileSize.QuadPart;
                        }

                    } else {

                        Status = STATUS_USER_MAPPED_FILE;
                        DbgBreak();
                        __leave;
                    }
                }

                if (NotifyFilter) {
                    SetFlag(FileObject->Flags, FO_FILE_MODIFIED);

                    Ext2SaveInode( IrpContext,
                                   Vcb,
                                   Fcb->Ext2Mcb->Inode,
                                   Fcb->Inode );

                    if (NT_SUCCESS(Status)) {
                        CcSetFileSizes(FileObject, (PCC_FILE_SIZES)(&(Fcb->Header.AllocationSize)));
                        SetFlag(FileObject->Flags, FO_FILE_MODIFIED);
                    }
                }

                Ext2Print((DBG_TRACE, "Ext2SetInformation: AllocatieonSize = %I64xh Allocated = %I64xh status = %xh\n",
                                       FAI->AllocationSize.QuadPart, Fcb->Header.AllocationSize.QuadPart, Status));

            }

            break;

        case FileEndOfFileInformation:
            {
                PFILE_END_OF_FILE_INFORMATION FEOFI = (PFILE_END_OF_FILE_INFORMATION) Buffer;

                if (IsDirectory(Fcb)) {
                    Status = STATUS_INVALID_DEVICE_REQUEST;
                    DbgBreak();
                    __leave;
                }

                if ((FileObject->SectionObjectPointer->DataSectionObject != NULL) &&
                    (FileObject->SectionObjectPointer->SharedCacheMap == NULL) &&
                    !FlagOn(Irp->Flags, IRP_PAGING_IO)) {

                    ASSERT( !FlagOn( FileObject->Flags, FO_CLEANUP_COMPLETE ) );

                    CcInitializeCacheMap( 
                            FileObject,
                            (PCC_FILE_SIZES)&(Fcb->Header.AllocationSize),
                            FALSE,
                            &(Ext2Global->CacheManagerNoOpCallbacks),
                            Fcb );

                    CacheInitialized = TRUE;
                }

                if (IoStackLocation->Parameters.SetFile.AdvanceOnly) {

                    LARGE_INTEGER   entrySize, EndOfFile;

                    Status = STATUS_SUCCESS;
                    if (IsFlagOn(Fcb->Flags, FCB_DELETE_PENDING)) {
                        __leave;
                    }

                    if (FEOFI->EndOfFile.QuadPart > Fcb->Header.FileSize.QuadPart) {
                        EndOfFile.QuadPart = Fcb->Header.FileSize.QuadPart;
                    } else {
                        EndOfFile.QuadPart = FEOFI->EndOfFile.QuadPart;
                    }

                    entrySize.LowPart = Fcb->Inode->i_size;
                    if (S_ISREG(Fcb->Inode->i_mode)) {
                        entrySize.HighPart = (LONG) (Fcb->Inode->i_size_high);
                    } else {
                        entrySize.HighPart = 0;
                    }

                    if (entrySize.QuadPart < EndOfFile.QuadPart) {
                        Fcb->Ext2Mcb->FileSize = EndOfFile;
                        Fcb->Inode->i_size = EndOfFile.LowPart;
                        if (S_ISREG(Fcb->Inode->i_mode)) {
                            Fcb->Inode->i_size_high = (__u32)(EndOfFile.HighPart);
                        }
                        Ext2SaveInode( IrpContext,
                                       Vcb,
                                       Fcb->Ext2Mcb->Inode,
                                       Fcb->Inode);
                        NotifyFilter = FILE_NOTIFY_CHANGE_SIZE;
                    }

                    __leave;
                }

                if (FEOFI->EndOfFile.QuadPart <= Fcb->Header.AllocationSize.QuadPart) {

                    Status = STATUS_SUCCESS;

                } else if (FEOFI->EndOfFile.QuadPart > Fcb->Header.AllocationSize.QuadPart) {

                    Status = Ext2ExpandFile(IrpContext, Vcb, Fcb, &(FEOFI->EndOfFile));
                }

                if (NT_SUCCESS(Status)) {

                    Fcb->Header.FileSize.QuadPart = Fcb->Ext2Mcb->FileSize.QuadPart = FEOFI->EndOfFile.QuadPart;
                    if (Fcb->Header.ValidDataLength.QuadPart > Fcb->Header.FileSize.QuadPart)
                        Fcb->Header.ValidDataLength.QuadPart = Fcb->Header.FileSize.QuadPart;
  
                    Fcb->Inode->i_size = FEOFI->EndOfFile.LowPart;
                    if (S_ISREG(Fcb->Inode->i_mode)) {
                        Fcb->Inode->i_size_high = (__u32)(FEOFI->EndOfFile.HighPart);
                    }

                    Ext2SaveInode( IrpContext, Vcb, Fcb->Ext2Mcb->Inode, Fcb->Inode);
                    CcSetFileSizes(FileObject, (PCC_FILE_SIZES)(&(Fcb->Header.AllocationSize)));
                    SetFlag(FileObject->Flags, FO_FILE_MODIFIED);

                    NotifyFilter = FILE_NOTIFY_CHANGE_SIZE;
                }

                Ext2Print((DBG_TRACE, "Ext2SetInformation: FileSize = %I64xh RealSize = %I64xh status = %xh\n",
                                      FEOFI->EndOfFile.QuadPart, Fcb->Header.FileSize.QuadPart, Status));
            }

            break;

        case FileDispositionInformation:
            {
                PFILE_DISPOSITION_INFORMATION FDI = (PFILE_DISPOSITION_INFORMATION)Buffer;
                
                Status = Ext2SetDispositionInfo(IrpContext, Vcb, Fcb, FDI->DeleteFile);
                
                Ext2Print((DBG_ERROR, "Ext2SetInformation: SetDispositionInformation: DeleteFile=%d %wZ status = %xh\n",
                                      FDI->DeleteFile, &Fcb->Ext2Mcb->ShortName, Status));
            }

            break;

        case FileRenameInformation:
            {
                Status = Ext2SetRenameInfo(IrpContext, Vcb, Fcb);
            }

            break;

            //
            // This is the only set file information request supported on read
            // only file systems
            //
        case FilePositionInformation:
            {
                PFILE_POSITION_INFORMATION FilePositionInformation;
                
                if (Length < sizeof(FILE_POSITION_INFORMATION)) {
                    Status = STATUS_INVALID_PARAMETER;
                    __leave;
                }
                
                FilePositionInformation = (PFILE_POSITION_INFORMATION) Buffer;
                
                if ((FlagOn(FileObject->Flags, FO_NO_INTERMEDIATE_BUFFERING)) &&
                    (FilePositionInformation->CurrentByteOffset.LowPart &
                    DeviceObject->AlignmentRequirement) ) {
                    Status = STATUS_INVALID_PARAMETER;
                    __leave;
                }
                
                FileObject->CurrentByteOffset =
                    FilePositionInformation->CurrentByteOffset;
                
                Status = STATUS_SUCCESS;
                __leave;
            }

            break;

        case FileLinkInformation:

            Status = STATUS_INVALID_DEVICE_REQUEST;
            break;

        default:
            Ext2Print((DBG_ERROR, "Ext2SetInformation: invalid class: %d\n",
                              FileInformationClass));
            Status = STATUS_INVALID_PARAMETER;/* STATUS_INVALID_INFO_CLASS; */
        }

    } __finally {
        
        if (FcbPagingIoResourceAcquired) {
            ExReleaseResourceLite(&Fcb->PagingIoResource);
        }
        
        if (NT_SUCCESS(Status) && (NotifyFilter != 0)) {
            Ext2NotifyReportChange(
                        IrpContext,
                        Vcb,
                        Fcb,
                        NotifyFilter,
                        FILE_ACTION_MODIFIED );

        }

        if (FcbMainResourceAcquired) {
            ExReleaseResourceLite(&Fcb->MainResource);
        }
        
        if (CacheInitialized) {
            CcUninitializeCacheMap(FileObject, NULL, NULL);
        }

        if (!IrpContext->ExceptionInProgress) {
            if (Status == STATUS_PENDING) {
                Ext2QueueRequest(IrpContext);
            } else {
                Ext2CompleteIrpContext(IrpContext,  Status);
            }
        }
    }
    
    return Status;
}

NTSTATUS
Ext2ExpandFile( PEXT2_IRP_CONTEXT IrpContext, 
                PEXT2_VCB Vcb, PEXT2_FCB Fcb,
                PLARGE_INTEGER AllocationSize)
{
    ULONG    dwRet = 0;
    NTSTATUS Status = STATUS_SUCCESS;

    if (AllocationSize->QuadPart <= Fcb->Header.AllocationSize.QuadPart) {
        return Status;
    }

    if (((LONGLONG)SUPER_BLOCK->s_free_blocks_count) * Vcb->BlockSize <=
        (AllocationSize->QuadPart - Fcb->Header.AllocationSize.QuadPart)  ) {
        Ext2Print((DBG_ERROR, "Ext2ExpandFile: There is no enough disk space available.\n"));
        return STATUS_DISK_FULL;
    }

    ExAcquireResourceExclusiveLite(&Vcb->MetaResource, TRUE);
    while (NT_SUCCESS(Status) && (AllocationSize->QuadPart > Fcb->Header.AllocationSize.QuadPart)) {
        Status = Ext2ExpandInode(IrpContext, Vcb, Fcb, &dwRet);
    }
    ExReleaseResourceLite(&Vcb->MetaResource);
    return Status;
}

NTSTATUS
Ext2TruncateFile( PEXT2_IRP_CONTEXT IrpContext,
                  PEXT2_VCB Vcb, PEXT2_FCB Fcb,
                  PLARGE_INTEGER AllocationSize)
{
    NTSTATUS Status = STATUS_SUCCESS;

    ExAcquireResourceExclusiveLite(&Vcb->MetaResource, TRUE);
    while (NT_SUCCESS(Status) && (AllocationSize->QuadPart <
                    Fcb->Header.AllocationSize.QuadPart)) {
        Status= Ext2TruncateInode(IrpContext, Vcb, Fcb);
    }
    ExReleaseResourceLite(&Vcb->MetaResource);

    return Status;
}

NTSTATUS
Ext2SetDispositionInfo(
            PEXT2_IRP_CONTEXT IrpContext,
            PEXT2_VCB Vcb,
            PEXT2_FCB Fcb,
            BOOLEAN bDelete)
{
    PIRP    Irp = IrpContext->Irp;
    PIO_STACK_LOCATION IrpSp;

    IrpSp = IoGetCurrentIrpStackLocation(Irp);

    Ext2Print((DBG_INFO, "Ext2SetDispositionInfo: bDelete=%x\n", bDelete));
    
    if (bDelete) {

        Ext2Print((DBG_INFO, "Ext2SetDispositionInformation: MmFlushImageSection on %s.\n", 
                             Fcb->AnsiFileName.Buffer));

        if (!MmFlushImageSection( &Fcb->SectionObject,
                                  MmFlushForDelete )) {
            return STATUS_CANNOT_DELETE;
        }

        if (Fcb->Ext2Mcb->Inode == EXT2_ROOT_INO) {
            return STATUS_CANNOT_DELETE;
        }

        if (IsDirectory(Fcb)) {
            if (!Ext2IsDirectoryEmpty(Vcb, Fcb)) {
                return STATUS_DIRECTORY_NOT_EMPTY;
            }
        }

        SetFlag(Fcb->Flags, FCB_DELETE_PENDING);
        IrpSp->FileObject->DeletePending = TRUE;

        if (IsDirectory(Fcb)) {
            FsRtlNotifyFullChangeDirectory( Vcb->NotifySync,
                                            &Vcb->NotifyList,
                                            Fcb,
                                            NULL,
                                            FALSE,
                                            FALSE,
                                            0,
                                            NULL,
                                            NULL,
                                            NULL );
        }

    } else {

        ClearFlag(Fcb->Flags, FCB_DELETE_PENDING);
        IrpSp->FileObject->DeletePending = FALSE;
    }

    return STATUS_SUCCESS;
}  

NTSTATUS
Ext2SetRenameInfo(
            PEXT2_IRP_CONTEXT IrpContext,
            PEXT2_VCB         Vcb,
            PEXT2_FCB         Fcb       )
{
    PEXT2_FCB               TargetDcb;
    PEXT2_MCB               TargetMcb;

    PEXT2_MCB               Mcb;

    UNICODE_STRING          FileName;
    
    NTSTATUS                Status;

    PIRP                    Irp;
    PIO_STACK_LOCATION      IrpSp;

    PFILE_OBJECT            FileObject;
    PFILE_OBJECT            TargetObject;
    BOOLEAN                 ReplaceIfExists;

    BOOLEAN                 bMove = FALSE;

    PFILE_RENAME_INFORMATION    FRI;

    if (Fcb->Ext2Mcb->Inode == EXT2_ROOT_INO) {
        Status = STATUS_INVALID_PARAMETER;
        goto errorout;
    }

    Irp = IrpContext->Irp;
    IrpSp = IoGetCurrentIrpStackLocation(Irp);

    FileObject = IrpSp->FileObject;
    TargetObject = IrpSp->Parameters.SetFile.FileObject;
    ReplaceIfExists = IrpSp->Parameters.SetFile.ReplaceIfExists;

    FRI = (PFILE_RENAME_INFORMATION)Irp->AssociatedIrp.SystemBuffer;

    if (TargetObject == NULL) {

        UNICODE_STRING  NewName;

        NewName.Buffer = FRI->FileName;
        NewName.MaximumLength = NewName.Length = (USHORT)FRI->FileNameLength;

        while (NewName.Length > 0 && NewName.Buffer[NewName.Length/2 - 1] == L'\\') {
            NewName.Buffer[NewName.Length/2 - 1] = 0;
            NewName.Length -= 2;
        }

        while (NewName.Length > 0 && NewName.Buffer[NewName.Length/2 - 1] != L'\\') {
            NewName.Length -= 2;
        }

        NewName.Buffer = (USHORT *)((UCHAR *)NewName.Buffer + NewName.Length);
        NewName.Length = (USHORT)(FRI->FileNameLength - NewName.Length);

        FileName = NewName;

        TargetDcb = NULL;
        TargetMcb = Fcb->Ext2Mcb->Parent;
        
        if (FileName.Length >= EXT2_NAME_LEN*sizeof(USHORT)) {
            Status = STATUS_OBJECT_NAME_INVALID;
            goto errorout;
        }

    } else {

        TargetDcb = (PEXT2_FCB)(TargetObject->FsContext);

        if (!TargetDcb || TargetDcb->Vcb != Vcb) {

            DbgBreak();

            Status = STATUS_INVALID_PARAMETER;
            goto errorout;
        }

        TargetMcb = TargetDcb->Ext2Mcb;

        FileName = TargetObject->FileName;
    }

    if (FsRtlDoesNameContainWildCards(&FileName)) {
        Status = STATUS_OBJECT_NAME_INVALID;
        goto errorout;
    }

    if (TargetMcb->Inode == Fcb->Ext2Mcb->Parent->Inode) {
        if (FsRtlAreNamesEqual( &FileName,
                                &(Fcb->Ext2Mcb->ShortName),
                                FALSE,
                                NULL )) {
            Status = STATUS_SUCCESS;
            goto errorout;
        }
    } else {
        bMove = TRUE;
    }

    TargetDcb = TargetMcb->Ext2Fcb;

    if (!TargetDcb)
        TargetDcb = Ext2CreateFcbFromMcb(Vcb, TargetMcb);

    if ((TargetMcb->Inode != Fcb->Ext2Mcb->Parent->Inode) &&
        (Fcb->Ext2Mcb->Parent->Ext2Fcb == NULL)      ) {
        Ext2CreateFcbFromMcb(Vcb, Fcb->Ext2Mcb->Parent);
    }

    if (!TargetDcb || !(Fcb->Ext2Mcb->Parent->Ext2Fcb)) {
        Status = STATUS_UNSUCCESSFUL;

        goto errorout;
    }

    Mcb = NULL;
    Status = Ext2LookupFileName(
                Vcb,
                &FileName,
                TargetMcb,
                &Mcb); 

    if (NT_SUCCESS(Status)) {

        if ( (!ReplaceIfExists) ||
             (IsFlagOn(Mcb->FileAttr, FILE_ATTRIBUTE_DIRECTORY)) ||
             (IsFlagOn(Mcb->FileAttr, FILE_ATTRIBUTE_READONLY))) {
            Status = STATUS_OBJECT_NAME_COLLISION;
            goto errorout;
        }

        if (ReplaceIfExists) {
            DbgBreak();
            Status = STATUS_NOT_IMPLEMENTED;
            goto errorout;
        }
    }

    if (IsDirectory(Fcb)) {

        Status = Ext2RemoveEntry( IrpContext, Vcb, 
                                  Fcb->Ext2Mcb->Parent->Ext2Fcb,
                                  EXT2_FT_DIR,
                                  Fcb->Ext2Mcb->Inode );

        if (!NT_SUCCESS(Status)) {
            DbgBreak();

            goto errorout;
        }

        Status = Ext2AddEntry( IrpContext, Vcb, 
                               TargetDcb,
                               EXT2_FT_DIR,
                               Fcb->Ext2Mcb->Inode,
                               &FileName );

        if (!NT_SUCCESS(Status)) {

            DbgBreak();

            Ext2AddEntry(  IrpContext, Vcb, 
                           Fcb->Ext2Mcb->Parent->Ext2Fcb,
                           EXT2_FT_DIR,
                           Fcb->Ext2Mcb->Inode,
                           &Fcb->Ext2Mcb->ShortName );

            goto errorout;
        }

        if( !Ext2SaveInode( IrpContext,
                            Vcb, 
                            TargetMcb->Inode,
                            TargetDcb->Inode)) {
            Status = STATUS_UNSUCCESSFUL;

            DbgBreak();

            goto errorout;
        }

        if( !Ext2SaveInode( IrpContext,
                            Vcb, 
                            Fcb->Ext2Mcb->Parent->Inode,
                            Fcb->Ext2Mcb->Parent->Ext2Fcb->Inode)) {

            Status = STATUS_UNSUCCESSFUL;

            DbgBreak();

            goto errorout;
        }

        Status = Ext2SetParentEntry( IrpContext, Vcb, Fcb,
                                     Fcb->Ext2Mcb->Parent->Inode,
                                     TargetDcb->Ext2Mcb->Inode );


        if (!NT_SUCCESS(Status)) {
            DbgBreak();
            goto errorout;
        }

    } else {

        Status = Ext2RemoveEntry( IrpContext, Vcb,
                                  Fcb->Ext2Mcb->Parent->Ext2Fcb,
                                  EXT2_FT_REG_FILE,
                                  Fcb->Ext2Mcb->Inode );
        if (!NT_SUCCESS(Status)) {
            DbgBreak();
            goto errorout;
        }

        Status = Ext2AddEntry( IrpContext,
                               Vcb, TargetDcb,
                               EXT2_FT_REG_FILE,
                               Fcb->Ext2Mcb->Inode,
                               &FileName );

        if (!NT_SUCCESS(Status)) {

            DbgBreak();

            Ext2AddEntry(  IrpContext, Vcb, 
                           Fcb->Ext2Mcb->Parent->Ext2Fcb,
                           EXT2_FT_REG_FILE,
                           Fcb->Ext2Mcb->Inode,
                           &Fcb->Ext2Mcb->ShortName );

            goto errorout;
        }
    }

    if (NT_SUCCESS(Status)) {

        if (bMove) {
            Ext2NotifyReportChange(
                       IrpContext,
                       Vcb,
                       Fcb,
                       (IsDirectory(Fcb) ?
                         FILE_NOTIFY_CHANGE_DIR_NAME :
                         FILE_NOTIFY_CHANGE_FILE_NAME ),
                       FILE_ACTION_REMOVED);

        } else {

            Ext2NotifyReportChange(
                       IrpContext,
                       Vcb,
                       Fcb,
                       (IsDirectory(Fcb) ?
                         FILE_NOTIFY_CHANGE_DIR_NAME :
                         FILE_NOTIFY_CHANGE_FILE_NAME ),
                       FILE_ACTION_RENAMED_OLD_NAME);

        }

        if (Fcb->Ext2Mcb->ShortName.MaximumLength < (FileName.Length + 2)) {

            if (Fcb->Ext2Mcb->ShortName.Buffer)
                ExFreePool(Fcb->Ext2Mcb->ShortName.Buffer);
            Fcb->Ext2Mcb->ShortName.Buffer = 
                ExAllocatePool(PagedPool, FileName.Length + 2);

            if (!Fcb->Ext2Mcb->ShortName.Buffer) {
                Ext2Print((DBG_ERROR, "Ex2SetRenameInfo: failed to allocate ShortName.\n"));
                Status = STATUS_INSUFFICIENT_RESOURCES;
                goto errorout;
            }

            Fcb->Ext2Mcb->ShortName.MaximumLength = FileName.Length + 2;
        }

        {
            RtlZeroMemory( Fcb->Ext2Mcb->ShortName.Buffer,
                           Fcb->Ext2Mcb->ShortName.MaximumLength);

            RtlCopyMemory( Fcb->Ext2Mcb->ShortName.Buffer,
                           FileName.Buffer, FileName.Length);

            Fcb->Ext2Mcb->ShortName.Length = FileName.Length;
        }
    
#if EXT2_DEBUG    

        Fcb->AnsiFileName.Length = (USHORT)
            Ext2UnicodeToOEMSize(Vcb, &FileName) + 1;

        if (Fcb->AnsiFileName.MaximumLength < FileName.Length) {
            ExFreePool(Fcb->AnsiFileName.Buffer);

            Fcb->AnsiFileName.Buffer = 
                ExAllocatePool(PagedPool, Fcb->AnsiFileName.Length + 1);

            if (!Fcb->AnsiFileName.Buffer)  {
                Ext2Print((DBG_ERROR, "Ex2SetRenameInfo: failed to allocate AnsiName.\n"));
                Status = STATUS_INSUFFICIENT_RESOURCES;
                goto errorout;
            }

            Fcb->AnsiFileName.MaximumLength = 
                           Fcb->AnsiFileName.Length + 1;
        }

        RtlZeroMemory(Fcb->AnsiFileName.Buffer, Fcb->AnsiFileName.MaximumLength);

        Ext2UnicodeToOEM( Vcb, 
                          &(Fcb->AnsiFileName),
                          &FileName          );
#endif

        Ext2DeleteMcbNode(Vcb, Fcb->Ext2Mcb->Parent, Fcb->Ext2Mcb);
        Ext2AddMcbNode(Vcb, TargetMcb, Fcb->Ext2Mcb);

        if (Fcb->LongName.Buffer) {
            ExFreePool(Fcb->LongName.Buffer);
            Fcb->LongName.Buffer = NULL;
        }

        if (bMove) {
            Ext2NotifyReportChange(
                       IrpContext,
                       Vcb,
                       Fcb,
                       (IsDirectory(Fcb) ?
                         FILE_NOTIFY_CHANGE_DIR_NAME :
                         FILE_NOTIFY_CHANGE_FILE_NAME ),
                       FILE_ACTION_ADDED);
        } else {
            Ext2NotifyReportChange(
                       IrpContext,
                       Vcb,
                       Fcb,
                       (IsDirectory(Fcb) ?
                         FILE_NOTIFY_CHANGE_DIR_NAME :
                         FILE_NOTIFY_CHANGE_FILE_NAME ),
                       FILE_ACTION_RENAMED_NEW_NAME  );

        }
    }

errorout:

    return Status;
}


NTSTATUS
Ext2DeleteFile(
        PEXT2_IRP_CONTEXT IrpContext,
        PEXT2_VCB Vcb,
        PEXT2_FCB Fcb )
{
    LARGE_INTEGER   AllocationSize;
    PEXT2_FCB       Dcb = NULL;

    NTSTATUS        Status = STATUS_SUCCESS;

    BOOLEAN         FcbPagingIoAcquired = FALSE;
    BOOLEAN         FcbResourceAcquired = FALSE;

    Ext2Print((DBG_INFO, "Ext2DeleteFile: File %S (%xh) will be deleted!\n",
                         Fcb->Ext2Mcb->ShortName.Buffer, Fcb->Ext2Mcb->Inode));

    if (IsFlagOn(Fcb->Flags, FCB_FILE_DELETED)) {
        return Status;
    }

    if (FlagOn(Fcb->Ext2Mcb->FileAttr, FILE_ATTRIBUTE_DIRECTORY)) {
        if (!Ext2IsDirectoryEmpty(Vcb, Fcb)) {
            ClearFlag(Fcb->Flags, FCB_DELETE_PENDING);
            
            return STATUS_DIRECTORY_NOT_EMPTY;
        }
    }

    Ext2Print((DBG_INFO, "Ext2DeleteFile: EXT2SB->S_FREE_BLOCKS = %xh .\n",
                         Vcb->SuperBlock->s_free_blocks_count));

    Status = STATUS_UNSUCCESSFUL;

    __try {

        if (Fcb->Ext2Mcb->Parent->Ext2Fcb) {

            Status = Ext2RemoveEntry(
                        IrpContext, Vcb, 
                        Fcb->Ext2Mcb->Parent->Ext2Fcb,
                        (FlagOn(Fcb->Ext2Mcb->FileAttr, FILE_ATTRIBUTE_DIRECTORY) ?
                         EXT2_FT_DIR : EXT2_FT_REG_FILE),
                        Fcb->Ext2Mcb->Inode);
        } else {

            Dcb = Ext2CreateFcbFromMcb(Vcb, Fcb->Ext2Mcb->Parent);
            if (Dcb) {
                Status = Ext2RemoveEntry(
                            IrpContext, Vcb, Dcb,
                            (FlagOn(Fcb->Ext2Mcb->FileAttr, FILE_ATTRIBUTE_DIRECTORY) ?
                             EXT2_FT_DIR : EXT2_FT_REG_FILE),
                            Fcb->Ext2Mcb->Inode);
            }
        }

        if (NT_SUCCESS(Status)) {

            FcbResourceAcquired = 
                ExAcquireResourceExclusiveLite(&Fcb->MainResource, TRUE);

            FcbPagingIoAcquired = 
                ExAcquireResourceExclusiveLite(&Fcb->PagingIoResource, TRUE);

            /* update i_links_count */
            Fcb->Inode->i_links_count--;

            if (IsDirectory(Fcb)) {

                if (Fcb->Inode->i_links_count >= 2) {
                    Status = STATUS_CANNOT_DELETE;
                }
            } else {
                if (Fcb->Inode->i_links_count > 0) {
                    Status = STATUS_CANNOT_DELETE;
                }
            }

            if (!NT_SUCCESS(Status)) {
                __leave;
            }

            /* truncate file size */
            AllocationSize.QuadPart = (LONGLONG)0;

            Status = Ext2TruncateFile(IrpContext, Vcb, Fcb, &AllocationSize);

            /* Update the inode's data length . It should be ZERO if succeeds. */

            if (Fcb->Header.FileSize.QuadPart > Fcb->Header.AllocationSize.QuadPart) {

                Fcb->Ext2Mcb->FileSize.QuadPart = Fcb->Header.AllocationSize.QuadPart;
                Fcb->Header.FileSize.QuadPart = Fcb->Header.AllocationSize.QuadPart;
                Fcb->Inode->i_size = Fcb->Header.AllocationSize.LowPart;
                if (S_ISREG(Fcb->Inode->i_mode)) {
                    Fcb->Inode->i_size_high = (__u32)(Fcb->Header.AllocationSize.HighPart);
                }
                if (Fcb->Header.ValidDataLength.QuadPart > Fcb->Header.FileSize.QuadPart) {
                    Fcb->Header.ValidDataLength.QuadPart = Fcb->Header.FileSize.QuadPart;
                }
            }
    
            {
                LARGE_INTEGER   SysTime;
                KeQuerySystemTime(&SysTime);

                Fcb->Inode->i_links_count = 0;
                Fcb->Inode->i_dtime = Ext2LinuxTime(SysTime);

                Ext2SaveInode(IrpContext, Vcb, Fcb->Ext2Mcb->Inode, Fcb->Inode);
            }

            if (IsDirectory(Fcb)) {
                Ext2FreeInode(IrpContext, Vcb, Fcb->Ext2Mcb->Inode, EXT2_FT_DIR);
            } else {
                Ext2FreeInode(IrpContext, Vcb, Fcb->Ext2Mcb->Inode, EXT2_FT_REG_FILE);
            }

            SetFlag(Fcb->Flags, FCB_FILE_DELETED);
            Ext2DeleteMcbNode(Vcb, Fcb->Ext2Mcb->Parent, Fcb->Ext2Mcb);
        }

    } __finally {

        if (FcbPagingIoAcquired) {
            ExReleaseResourceLite(&Fcb->PagingIoResource);
        }

        if (FcbResourceAcquired) {
            ExReleaseResourceLite(&Fcb->MainResource);
        }
    }

    Ext2Print((DBG_INFO, "Ext2DeleteFile: Succeed... EXT2SB->S_FREE_BLOCKS = %xh .\n",
                         Vcb->SuperBlock->s_free_blocks_count));

    return Status;
}
